#include <pthread.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <linux/input.h>

#include "hal.h"
#include "fce.h"

#define MAX_KEYCODE	64
#define LOCAL_SIZE	(480 * 512 * 2)
#define FBMAP_SIZE	(480 * 800 * 2)
#define LINE_PIXEL	480

char key_table[MAX_KEYCODE] = {0};
unsigned short *local;
unsigned short *framebuffer = NULL;
unsigned short pal_table[64];
unsigned int framecount = 0;

void nes_set_bg_color(int c)
{
	if (framecount % 2)
		return;

	int size = LOCAL_SIZE / sizeof(int), i;
	int colorgroup = (pal_table[c] << 16) | pal_table[c];
	for (i = 0; i < size; i++)
		((int *)local)[i] = colorgroup;
}

void nes_flush_buf(PixelBuf *buf)
{
	if (framecount % 2)
		return;

	for (int i = 0; i < buf->size; i++) {
	    Pixel *p = &buf->buf[i];
	    unsigned short c = pal_table[p->c];
	    int pos;

	    pos = p->x * 2 * LINE_PIXEL + 478 - p->y * 2;
	    local[pos] = c;
	    local[pos + 1] = c;

	    pos += LINE_PIXEL;
	    local[pos] = c;
	    local[pos + 1] = c;
	}
}

void nes_flip_display()
{
	if (framecount++ % 2)
		return;

	// directly write to middle of framebuffer.
	memcpy(framebuffer + 144 * 480, local, LOCAL_SIZE);
}

void* keyboard_thread(void *arg)
{
	struct input_event ie;
	int kf;

	kf = open("/dev/input/event0", O_RDONLY);
	if (kf < 0) {
		printf("keyboard is not exist, input ignore.\n");
		return NULL;
	}

	while (1) {
		int size;
		size = read(kf, &ie, sizeof(struct input_event));
		if (size != sizeof(struct input_event))
			break;

		if (ie.type != EV_KEY || ie.code >= MAX_KEYCODE)
			continue;

		key_table[ie.code] = ie.value;
		if (ie.code == KEY_ESC)
			break;
	};

	close(kf);
	return arg;
}

void nes_hal_init()
{
	// create thread to listen to keyboard event.
	pthread_t tid;
	pthread_create(&tid, NULL, keyboard_thread, NULL);

	// create local frame buffer.
	local = malloc(LOCAL_SIZE);
	memset(local, 0, LOCAL_SIZE);

	// map framebuffer to local, so we can access it directly.
	int fd = open("/dev/fb0", O_RDWR);
	if (fd <= 0) {
		printf("frame buffer is not exist, exit.\n");
		exit(-1);
	}

	framebuffer = mmap(NULL, FBMAP_SIZE, PROT_WRITE, MAP_SHARED, fd, 0);
	if (framebuffer == NULL) {
		printf("can not map to frame buffer, exit.\n");
		exit(-1);
	}

	// convert palette 24bit to 16bit.
	for (int i = 0; i < 64; i++) {
		unsigned short hb, lb;
		hb = ((palette[i].r >> 3) | ((palette[i].g << 3) & 0xe0));
		lb = ((palette[i].b & 0xf8) | (palette[i].g >> 5));
		pal_table[i] = (hb << 8) | lb;
	}
}

void wait_for_frame()
{
	// avoid 100% CPU usage and make it as near as possible to the real NES
	// timing. :)
	usleep(2200);
}

int nes_key_state(int b)
{
	// we need to check keyboard, if it is time to quit.
	if (key_table[KEY_ESC])
		exit(0);

	switch (b) {
	case 0: // On / Off
		return 1;
	case 1: // A
		return key_table[KEY_K];
	case 2: // B
		return key_table[KEY_J];
	case 3: // SELECT
		return key_table[KEY_U];
	case 4: // START
		return key_table[KEY_I];
	case 5: // UP
		return key_table[KEY_W];
	case 6: // DOWN
		return key_table[KEY_S];
	case 7: // LEFT
		return key_table[KEY_A];
	case 8: // RIGHT
		return key_table[KEY_D];
	default:
		return 1;
	}
}

